@@ -21,7 +21,7 @@
 // Copyright (C) 2006, 2007, 2010, 2011 Carlos Garcia Campos <carlosgc@gnome.org>
 // Copyright (C) 2007 Koji Otani <sho@bbr.jp>
 // Copyright (C) 2008, 2009 Chris Wilson <chris@chris-wilson.co.uk>
-// Copyright (C) 2008, 2012, 2014, 2016, 2017, 2022, 2023 Adrian Johnson <ajohnson@redneon.com>
+// Copyright (C) 2008, 2012, 2014, 2016, 2017, 2022-2024 Adrian Johnson <ajohnson@redneon.com>
 // Copyright (C) 2009 Darren Kenny <darren.kenny@sun.com>
 // Copyright (C) 2010 Suzuki Toshiya <mpsuzuki@hiroshima-u.ac.jp>
 // Copyright (C) 2010 Jan Kümmel <jan+freedesktop@snorc.org>
@@ -33,6 +33,10 @@
 // Copyright (C) 2020 Michal <sudolskym@gmail.com>
 // Copyright (C) 2021, 2022 Oliver Sander <oliver.sander@tu-dresden.de>
 // Copyright (C) 2022 Marcel Fabian Krüger <tex@2krueger.de>
+// Copyright (C) 2023 Pablo Correa Gómez <ablocorrea@hotmail.com>
+// Copyright (C) 2023 Frederic Germain <frederic.germain@gmail.com>
+// Copyright (C) 2023 Ilia Kats <ilia-kats@gmx.net>
+// Copyright (C) 2024 Vincent Lefevre <vincent@vinc17.net>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -163,7 +167,7 @@ CairoFreeTypeFont::CairoFreeTypeFont(Ref refA, cairo_font_face_t *cairo_font_fac
 CairoFreeTypeFont::~CairoFreeTypeFont() { }
 
 // Create a cairo_font_face_t for the given font filename OR font data.
-static std::optional<FreeTypeFontFace> createFreeTypeFontFace(FT_Library lib, const std::string &filename, std::vector<unsigned char> &&font_data)
+std::optional<FreeTypeFontFace> CairoFreeTypeFont::createFreeTypeFontFace(FT_Library lib, const std::string &filename, std::vector<unsigned char> &&font_data)
 {
     FreeTypeFontResource *resource = new FreeTypeFontResource;
     FreeTypeFontFace font_face;
@@ -194,17 +198,6 @@ static std::optional<FreeTypeFontFace> createFreeTypeFontFace(FT_Library lib, co
     return font_face;
 }
 
-// Create a cairo_font_face_t for the given font filename OR font data. First checks if external font
-// is in the cache.
-std::optional<FreeTypeFontFace> CairoFreeTypeFont::getFreeTypeFontFace(CairoFontEngine *fontEngine, FT_Library lib, const std::string &filename, std::vector<unsigned char> &&font_data)
-{
-    if (font_data.empty()) {
-        return fontEngine->getExternalFontFace(lib, filename);
-    }
-
-    return createFreeTypeFontFace(lib, filename, std::move(font_data));
-}
-
 CairoFreeTypeFont *CairoFreeTypeFont::create(const std::shared_ptr<GfxFont> &gfxFont, XRef *xref, FT_Library lib, CairoFontEngine *fontEngine, bool useCIDs)
 {
     std::string fileName;
@@ -247,7 +240,7 @@ CairoFreeTypeFont *CairoFreeTypeFont::create(const std::shared_ptr<GfxFont> &gfx
     case fontType1:
     case fontType1C:
     case fontType1COT:
-        font_face = getFreeTypeFontFace(fontEngine, lib, fileName, std::move(font_data));
+        font_face = createFreeTypeFontFace(lib, fileName, std::move(font_data));
         if (!font_face) {
             error(errSyntaxError, -1, "could not create type1 face");
             goto err2;
@@ -308,7 +301,7 @@ CairoFreeTypeFont *CairoFreeTypeFont::create(const std::shared_ptr<GfxFont> &gfx
             ff = FoFiTrueType::load(fileName.c_str());
         }
         if (!ff) {
-            error(errSyntaxError, -1, "failed to load truetype font\n");
+            error(errSyntaxError, -1, "failed to load truetype font");
             goto err2;
         }
         /* This might be set already for the CIDType2 case */
@@ -318,9 +311,9 @@ CairoFreeTypeFont *CairoFreeTypeFont::create(const std::shared_ptr<GfxFont> &gfx
             codeToGID.insert(codeToGID.begin(), src, src + 256);
             gfree(src);
         }
-        font_face = getFreeTypeFontFace(fontEngine, lib, fileName, std::move(font_data));
+        font_face = createFreeTypeFontFace(lib, fileName, std::move(font_data));
         if (!font_face) {
-            error(errSyntaxError, -1, "could not create truetype face\n");
+            error(errSyntaxError, -1, "could not create truetype face");
             goto err2;
         }
         break;
@@ -342,9 +335,9 @@ CairoFreeTypeFont *CairoFreeTypeFont::create(const std::shared_ptr<GfxFont> &gfx
             }
         }
 
-        font_face = getFreeTypeFontFace(fontEngine, lib, fileName, std::move(font_data));
+        font_face = createFreeTypeFontFace(lib, fileName, std::move(font_data));
         if (!font_face) {
-            error(errSyntaxError, -1, "could not create cid face\n");
+            error(errSyntaxError, -1, "could not create cid face");
             goto err2;
         }
         break;
@@ -377,9 +370,9 @@ CairoFreeTypeFont *CairoFreeTypeFont::create(const std::shared_ptr<GfxFont> &gfx
                 }
             }
         }
-        font_face = getFreeTypeFontFace(fontEngine, lib, fileName, std::move(font_data));
+        font_face = createFreeTypeFontFace(lib, fileName, std::move(font_data));
         if (!font_face) {
-            error(errSyntaxError, -1, "could not create cid (OT) face\n");
+            error(errSyntaxError, -1, "could not create cid (OT) face");
             goto err2;
         }
         break;
@@ -405,18 +398,20 @@ static const cairo_user_data_key_t type3_font_key = { 0 };
 
 typedef struct _type3_font_info
 {
-    _type3_font_info(const std::shared_ptr<GfxFont> &fontA, PDFDoc *docA, CairoFontEngine *fontEngineA, bool printingA, XRef *xrefA) : font(fontA), doc(docA), fontEngine(fontEngineA), printing(printingA), xref(xrefA) { }
+    _type3_font_info(const std::shared_ptr<GfxFont> &fontA, PDFDoc *docA, CairoFontEngine *fontEngineA, CairoOutputDev *outputDevA, Gfx *gfxA) : font(fontA), doc(docA), fontEngine(fontEngineA), outputDev(outputDevA), gfx(gfxA) { }
 
     std::shared_ptr<GfxFont> font;
     PDFDoc *doc;
     CairoFontEngine *fontEngine;
-    bool printing;
-    XRef *xref;
+    CairoOutputDev *outputDev;
+    Gfx *gfx;
 } type3_font_info_t;
 
 static void _free_type3_font_info(void *closure)
 {
     type3_font_info_t *info = (type3_font_info_t *)closure;
+    delete info->gfx;
+    delete info->outputDev;
     delete info;
 }
 
@@ -439,17 +434,15 @@ static cairo_status_t _render_type3_glyph(cairo_scaled_font_t *scaled_font, unsi
 {
     Dict *charProcs;
     Object charProc;
-    CairoOutputDev *output_dev;
     cairo_matrix_t matrix, invert_y_axis;
     const double *mat;
     double wx, wy;
-    PDFRectangle box;
     type3_font_info_t *info;
+    Gfx *gfx;
     cairo_status_t status;
 
     info = (type3_font_info_t *)cairo_font_face_get_user_data(cairo_scaled_font_get_font_face(scaled_font), &type3_font_key);
 
-    Dict *resDict = std::static_pointer_cast<Gfx8BitFont>(info->font)->getResources();
     charProcs = std::static_pointer_cast<Gfx8BitFont>(info->font)->getCharProcs();
     if (!charProcs) {
         return CAIRO_STATUS_USER_FONT_ERROR;
@@ -470,16 +463,16 @@ static cairo_status_t _render_type3_glyph(cairo_scaled_font_t *scaled_font, unsi
     cairo_matrix_multiply(&matrix, &matrix, &invert_y_axis);
     cairo_transform(cr, &matrix);
 
-    output_dev = new CairoOutputDev();
+#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 18, 0)
+    cairo_set_source(cr, cairo_user_scaled_font_get_foreground_marker(scaled_font));
+#endif
+
+    CairoOutputDev *output_dev = info->outputDev;
     output_dev->setCairo(cr);
-    output_dev->setPrinting(info->printing);
+
+    gfx = info->gfx;
+    gfx->saveState();
 
-    mat = info->font->getFontBBox();
-    box.x1 = mat[0];
-    box.y1 = mat[1];
-    box.x2 = mat[2];
-    box.y2 = mat[3];
-    auto gfx = std::make_unique<Gfx>(info->doc, output_dev, resDict, &box, nullptr);
     output_dev->startDoc(info->doc, info->fontEngine);
     output_dev->startType3Render(gfx->getState(), gfx->getXRef());
     output_dev->setType3RenderType(color ? CairoOutputDev::Type3RenderColor : CairoOutputDev::Type3RenderMask);
@@ -510,6 +503,7 @@ static cairo_status_t _render_type3_glyph(cairo_scaled_font_t *scaled_font, unsi
         metrics->width = bbox[2] - bbox[0];
         metrics->height = bbox[3] - bbox[1];
     }
+    gfx->restoreState();
 
     status = CAIRO_STATUS_SUCCESS;
 
@@ -520,12 +514,10 @@ static cairo_status_t _render_type3_glyph(cairo_scaled_font_t *scaled_font, unsi
         status = CAIRO_STATUS_USER_FONT_NOT_IMPLEMENTED;
     }
 
-    delete output_dev;
-
     return status;
 }
 
-#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 17, 6)
+#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 18, 0)
 static cairo_status_t _render_type3_color_glyph(cairo_scaled_font_t *scaled_font, unsigned long glyph, cairo_t *cr, cairo_text_extents_t *metrics)
 {
     return _render_type3_glyph(scaled_font, glyph, cr, metrics, true);
@@ -541,20 +533,33 @@ CairoType3Font *CairoType3Font::create(const std::shared_ptr<GfxFont> &gfxFont,
 {
     std::vector<int> codeToGID;
     char *name;
+    const double *mat;
 
     Dict *charProcs = std::static_pointer_cast<Gfx8BitFont>(gfxFont)->getCharProcs();
     Ref ref = *gfxFont->getID();
     cairo_font_face_t *font_face = cairo_user_font_face_create();
     cairo_user_font_face_set_init_func(font_face, _init_type3_glyph);
-#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 17, 6)
+#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 18, 0)
     // When both callbacks are set, Cairo will call the color glyph
     // callback first.  If that returns NOT_IMPLEMENTED, Cairo will
     // then call the non-color glyph callback.
     cairo_user_font_face_set_render_color_glyph_func(font_face, _render_type3_color_glyph);
 #endif
     cairo_user_font_face_set_render_glyph_func(font_face, _render_type3_noncolor_glyph);
-    type3_font_info_t *info = new type3_font_info_t(gfxFont, doc, fontEngine, printing, xref);
 
+    CairoOutputDev *output_dev = new CairoOutputDev();
+    output_dev->setPrinting(printing);
+
+    Dict *resDict = std::static_pointer_cast<Gfx8BitFont>(gfxFont)->getResources();
+    mat = gfxFont->getFontBBox();
+    PDFRectangle box;
+    box.x1 = mat[0];
+    box.y1 = mat[1];
+    box.x2 = mat[2];
+    box.y2 = mat[3];
+    Gfx *gfx = new Gfx(doc, output_dev, resDict, &box, nullptr);
+
+    type3_font_info_t *info = new type3_font_info_t(gfxFont, doc, fontEngine, output_dev, gfx);
     cairo_font_face_set_user_data(font_face, &type3_font_key, (void *)info, _free_type3_font_info);
 
     char **enc = std::static_pointer_cast<Gfx8BitFont>(gfxFont)->getEncoding();
@@ -586,9 +591,6 @@ bool CairoType3Font::matches(Ref &other, bool printingA)
 // CairoFontEngine
 //------------------------------------------------------------------------
 
-std::unordered_map<std::string, FreeTypeFontFace> CairoFontEngine::fontFileCache;
-std::recursive_mutex CairoFontEngine::fontFileCacheMutex;
-
 CairoFontEngine::CairoFontEngine(FT_Library libA)
 {
     lib = libA;
@@ -637,33 +639,3 @@ std::shared_ptr<CairoFont> CairoFontEngine::getFont(const std::shared_ptr<GfxFon
     }
     return font;
 }
-
-std::optional<FreeTypeFontFace> CairoFontEngine::getExternalFontFace(FT_Library ftlib, const std::string &filename)
-{
-    std::scoped_lock lock(fontFileCacheMutex);
-
-    auto it = fontFileCache.find(filename);
-    if (it != fontFileCache.end()) {
-        FreeTypeFontFace font = it->second;
-        cairo_font_face_reference(font.cairo_font_face);
-        return font;
-    }
-
-    std::optional<FreeTypeFontFace> font_face = createFreeTypeFontFace(ftlib, filename, {});
-    if (font_face) {
-        cairo_font_face_reference(font_face->cairo_font_face);
-        fontFileCache[filename] = *font_face;
-    }
-
-    it = fontFileCache.begin();
-    while (it != fontFileCache.end()) {
-        if (cairo_font_face_get_reference_count(it->second.cairo_font_face) == 1) {
-            cairo_font_face_destroy(it->second.cairo_font_face);
-            it = fontFileCache.erase(it);
-        } else {
-            ++it;
-        }
-    }
-
-    return font_face;
-}
