@@ -14,7 +14,7 @@
 // under GPL version 2 or later
 //
 // Copyright (C) 2005, 2006, 2008 Brad Hards <bradh@frogmouth.net>
-// Copyright (C) 2005, 2007-2009, 2011-2023 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2005, 2007-2009, 2011-2024 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2008 Julien Rebetez <julienr@svn.gnome.org>
 // Copyright (C) 2008, 2010 Pino Toscano <pino@kde.org>
 // Copyright (C) 2008, 2010, 2011 Carlos Garcia Campos <carlosgc@gnome.org>
@@ -46,7 +46,7 @@
 // Copyright (C) 2020 Nelson Benítez León <nbenitezl@gmail.com>
 // Copyright (C) 2020 Thorsten Behrens <Thorsten.Behrens@CIB.de>
 // Copyright (C) 2020 Adam Sampson <ats@offog.org>
-// Copyright (C) 2021, 2022 Oliver Sander <oliver.sander@tu-dresden.de>
+// Copyright (C) 2021-2024 Oliver Sander <oliver.sander@tu-dresden.de>
 // Copyright (C) 2021 Mahmoud Khalil <mahmoudkhalil11@gmail.com>
 // Copyright (C) 2021 RM <rm+git@arcsin.org>
 // Copyright (C) 2021 Georgiy Sgibnev <georgiy@sgibnev.com>. Work sponsored by lab50.net.
@@ -54,7 +54,8 @@
 // Copyright (C) 2022 Felix Jung <fxjung@posteo.de>
 // Copyright (C) 2022 crt <chluo@cse.cuhk.edu.hk>
 // Copyright (C) 2022 Erich E. Hoover <erich.e.hoover@gmail.com>
-// Copyright (C) 2023 g10 Code GmbH, Author: Sune Stolborg Vuorela <sune@vuorela.dk>
+// Copyright (C) 2023, 2024 g10 Code GmbH, Author: Sune Stolborg Vuorela <sune@vuorela.dk>
+// Copyright (C) 2024 Vincent Lefevre <vincent@vinc17.net>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -64,6 +65,7 @@
 #include <config.h>
 #include <poppler-config.h>
 
+#include <array>
 #include <cctype>
 #include <clocale>
 #include <cstdio>
@@ -718,7 +720,7 @@ bool PDFDoc::isLinearized(bool tryingToReconstruct)
 
 void PDFDoc::setDocInfoStringEntry(const char *key, GooString *value)
 {
-    bool removeEntry = !value || value->getLength() == 0 || value->hasJustUnicodeMarker();
+    bool removeEntry = !value || value->getLength() == 0 || (value->toStr() == unicodeByteOrderMark);
     if (removeEntry) {
         delete value;
     }
@@ -1052,16 +1054,19 @@ int PDFDoc::saveWithoutChangesAs(const GooString &name)
 
 int PDFDoc::saveWithoutChangesAs(OutStream *outStr)
 {
-    int c;
-
     if (file && file->modificationTimeChangedSinceOpen()) {
         return errFileChangedSinceOpen;
     }
 
     BaseStream *copyStr = str->copy();
     copyStr->reset();
-    while ((c = copyStr->getChar()) != EOF) {
-        outStr->put(c);
+    while (copyStr->lookChar() != EOF) {
+        std::array<unsigned char, 4096> array;
+        size_t size = copyStr->doGetChars(array.size(), array.data());
+        auto sizeWritten = outStr->write(std::span(array.data(), size));
+        if (size != sizeWritten) {
+            return errFileIO;
+        }
     }
     copyStr->close();
     delete copyStr;
@@ -1071,13 +1076,16 @@ int PDFDoc::saveWithoutChangesAs(OutStream *outStr)
 
 void PDFDoc::saveIncrementalUpdate(OutStream *outStr)
 {
-    XRef *uxref;
-    int c;
     // copy the original file
     BaseStream *copyStr = str->copy();
     copyStr->reset();
-    while ((c = copyStr->getChar()) != EOF) {
-        outStr->put(c);
+    while (copyStr->lookChar() != EOF) {
+        std::array<unsigned char, 4096> array;
+        size_t size = copyStr->doGetChars(array.size(), array.data());
+        auto sizeWritten = outStr->write(std::span(array.data(), size));
+        if (size != sizeWritten) {
+            // Write error of some sort
+        }
     }
     copyStr->close();
     delete copyStr;
@@ -1087,7 +1095,7 @@ void PDFDoc::saveIncrementalUpdate(OutStream *outStr)
     int keyLength;
     xref->getEncryptionParameters(&fileKey, &encAlgorithm, &keyLength);
 
-    uxref = new XRef();
+    XRef *uxref = new XRef();
     uxref->add(0, 65535, 0, false);
     xref->lock();
     for (int i = 0; i < xref->getNumObjects(); i++) {
@@ -1205,6 +1213,23 @@ void PDFDoc::saveCompleteRewrite(OutStream *outStr)
     delete uxref;
 }
 
+std::string PDFDoc::sanitizedName(const std::string &name)
+{
+    std::string sanitizedName;
+
+    for (const auto c : name) {
+        if (c <= (char)0x20 || c >= (char)0x7f || c == ' ' || c == '(' || c == ')' || c == '<' || c == '>' || c == '[' || c == ']' || c == '{' || c == '}' || c == '/' || c == '%' || c == '#') {
+            char buf[8];
+            sprintf(buf, "#%02x", c & 0xff);
+            sanitizedName.append(buf);
+        } else {
+            sanitizedName.push_back(c);
+        }
+    }
+
+    return sanitizedName;
+}
+
 void PDFDoc::writeDictionary(Dict *dict, OutStream *outStr, XRef *xRef, unsigned int numOffset, unsigned char *fileKey, CryptAlgorithm encAlgorithm, int keyLength, Ref ref, std::set<Dict *> *alreadyWrittenDicts)
 {
     bool deleteSet = false;
@@ -1226,9 +1251,7 @@ void PDFDoc::writeDictionary(Dict *dict, OutStream *outStr, XRef *xRef, unsigned
     outStr->printf("<<");
     for (int i = 0; i < dict->getLength(); i++) {
         GooString keyName(dict->getKey(i));
-        GooString *keyNameToPrint = keyName.sanitizedName();
-        outStr->printf("/%s ", keyNameToPrint->c_str());
-        delete keyNameToPrint;
+        outStr->printf("/%s ", sanitizedName(keyName.toStr()).c_str());
         Object obj1 = dict->getValNF(i).copy();
         writeObject(&obj1, outStr, xRef, numOffset, fileKey, encAlgorithm, keyLength, ref, alreadyWrittenDicts);
     }
@@ -1296,7 +1319,7 @@ void PDFDoc::writeString(const GooString *s, OutStream *outStr, const unsigned c
     }
 
     // Write data
-    if (s->hasUnicodeMarker()) {
+    if (hasUnicodeByteOrderMark(s->toStr())) {
         // unicode string don't necessary end with \0
         const char *c = s->c_str();
         std::stringstream stream;
@@ -1376,9 +1399,7 @@ void PDFDoc::writeObject(Object *obj, OutStream *outStr, XRef *xRef, unsigned in
     }
     case objName: {
         GooString name(obj->getName());
-        GooString *nameToPrint = name.sanitizedName();
-        outStr->printf("/%s ", nameToPrint->c_str());
-        delete nameToPrint;
+        outStr->printf("/%s ", sanitizedName(name.toStr()).c_str());
         break;
     }
     case objNull:
@@ -1504,7 +1525,7 @@ void PDFDoc::writeObject(Object *obj, OutStream *outStr, XRef *xRef, unsigned in
         outStr->printf("none\r\n");
         break;
     default:
-        error(errUnimplemented, -1, "Unhandled objType : {0:d}, please report a bug with a testcase\r\n", obj->getType());
+        error(errUnimplemented, -1, "Unhandled objType : {0:d}, please report a bug with a testcase", obj->getType());
         break;
     }
 }
@@ -2207,6 +2228,8 @@ bool PDFDoc::sign(const std::string &saveFilename, const std::string &certNickna
     catalog->addFormToAcroForm(ref);
     // say that there a now signatures and that we should append only
     catalog->getAcroForm()->dictSet("SigFlags", Object(3));
+    catalog->setAcroFormModified();
+
     form->ensureFontsForAllCharacters(&signatureText, pdfFontName);
     form->ensureFontsForAllCharacters(&signatureTextLeft, pdfFontName);
 
